@@ -20,13 +20,9 @@
 
   Header objects:
   
-  Script Command (3.6)
   Marker (3.7)
   Bitrate Mutual Exclusion (3.8)
   Content Branding (3.13)
-  Content Encryption (3.14)
-  Extended Content Encryption (3.15)
-  Digital Signature (3.16)
   
   Header Extension objects:
   
@@ -144,6 +140,22 @@ get_asf_metadata(char *file, HV *info, HV *tags)
       DEBUG_TRACE("Stream_Bitrate_Properties\n");
       _parse_stream_bitrate_properties(&asf_buf, info, tags);
     }
+    else if ( IsEqualGUID(&tmp.ID, &ASF_Content_Encryption) ) {
+      DEBUG_TRACE("Content_Encryption\n");
+      _parse_content_encryption(&asf_buf, info, tags);
+    }
+    else if ( IsEqualGUID(&tmp.ID, &ASF_Extended_Content_Encryption) ) {
+      DEBUG_TRACE("Skipping Extended_Content_Encryption\n");
+      buffer_consume(&asf_buf, tmp.size - 24);
+    }
+    else if ( IsEqualGUID(&tmp.ID, &ASF_Script_Command) ) {
+      DEBUG_TRACE("Script_Command\n");
+      _parse_script_command(&asf_buf, info, tags);
+    }
+    else if ( IsEqualGUID(&tmp.ID, &ASF_Digital_Signature) ) {
+      DEBUG_TRACE("Skipping Digital_Signature\n");
+      buffer_consume(&asf_buf, tmp.size - 24);
+    }
     else if ( IsEqualGUID(&tmp.ID, &ASF_Header_Extension) ) {
       DEBUG_TRACE("Header_Extension\n");
       if ( !_parse_header_extension(&asf_buf, tmp.size, info, tags) ) {
@@ -258,7 +270,6 @@ _parse_extended_content_description(Buffer *buf, HV *info, HV *tags)
     uint16_t value_len;
     SV *key = NULL;
     SV *value = NULL;
-    HV *picture;
     Buffer utf8_buf;
     
     name_len = buffer_get_short_le(buf);
@@ -280,49 +291,7 @@ _parse_extended_content_description(Buffer *buf, HV *info, HV *tags)
     else if (data_type == TYPE_BYTE) {
       // handle picture data, interestingly it is compatible with the ID3v2 APIC frame
       if ( !strcmp( SvPVX(key), "WM/Picture" ) ) {
-        char *tmp_ptr;
-        uint16_t mime_len = 2; // to handle double-null
-        uint16_t desc_len = 2;
-        uint32_t image_len;
-        SV *mime;
-        SV *desc;
-        
-        picture = newHV();
-        
-        my_hv_store( picture, "image_type", newSViv( buffer_get_char(buf) ) );
-
-        image_len = buffer_get_int_le(buf);
-        
-        // MIME type is a double-null-terminated UTF-16 string
-        tmp_ptr = buffer_ptr(buf);
-        while ( tmp_ptr[0] != '\0' || tmp_ptr[1] != '\0' ) {
-          mime_len += 2;
-          tmp_ptr += 2;
-        }
-        
-        buffer_get_utf16le_as_utf8(buf, &utf8_buf, mime_len);
-        mime = newSVpv( buffer_ptr(&utf8_buf), 0 );
-        sv_utf8_decode(mime);
-        my_hv_store( picture, "mime_type", mime );
-        buffer_free(&utf8_buf);
-        
-        // Description is a double-null-terminated UTF-16 string
-        tmp_ptr = buffer_ptr(buf);
-        while ( tmp_ptr[0] != '\0' || tmp_ptr[1] != '\0' ) {
-          desc_len += 2;
-          tmp_ptr += 2;
-        }
-        
-        buffer_get_utf16le_as_utf8(buf, &utf8_buf, desc_len);
-        desc = newSVpv( buffer_ptr(&utf8_buf), 0 );
-        sv_utf8_decode(desc);
-        my_hv_store( picture, "description", desc );
-        buffer_free(&utf8_buf);
-        
-        my_hv_store( picture, "image", newSVpvn( buffer_ptr(buf), image_len ) );
-        buffer_consume(buf, image_len);
-        
-        value = newRV_noinc( (SV *)picture );
+        value = _parse_picture(buf);
       }
       else {
         value = newSVpvn( buffer_ptr(buf), value_len );
@@ -514,7 +483,9 @@ _parse_stream_properties(Buffer *buf, HV *info, HV *tags)
   else if ( IsEqualGUID(&stream_type, &ASF_JFIF_Media) ) {
     _store_stream_info( stream_number, info, newSVpv("stream_type", 0), newSVpv("ASF_JFIF_Media", 0) );
     
-    // XXX: type-specific data (section 9.4.1)
+    // type-specific data
+    _store_stream_info( stream_number, info, newSVpv("width", 0), newSVuv( buffer_get_int_le(&type_data_buf) ) );
+    _store_stream_info( stream_number, info, newSVpv("height", 0), newSVuv( buffer_get_int_le(&type_data_buf) ) );
   }
   else if ( IsEqualGUID(&stream_type, &ASF_Degradable_JPEG_Media) ) {
     _store_stream_info( stream_number, info, newSVpv("stream_type", 0), newSVpv("ASF_Degradable_JPEG_Media", 0) );
@@ -540,7 +511,7 @@ _parse_stream_properties(Buffer *buf, HV *info, HV *tags)
   }
   
   _store_stream_info( stream_number, info, newSVpv("time_offset", 0), newSViv(time_offset) );
-  _store_stream_info( stream_number, info, newSVpv("encrypted", 0), newSViv( flags & 0x8000 ) );
+  _store_stream_info( stream_number, info, newSVpv("encrypted", 0), newSVuv( flags & 0x8000 ? 1 : 0 ) );
   
   buffer_free(&type_data_buf);
 }
@@ -961,8 +932,14 @@ _parse_metadata_library(Buffer *buf, HV *info, HV *tags)
       buffer_free(&utf8_buf);
     }
     else if (data_type == TYPE_BYTE) {
-      value = newSVpvn( buffer_ptr(buf), data_len );
-      buffer_consume(buf, data_len);
+      // handle picture data
+      if ( !strcmp( SvPVX(key), "WM/Picture" ) ) {
+        value = _parse_picture(buf);
+      }
+      else {
+        value = newSVpvn( buffer_ptr(buf), data_len );
+        buffer_consume(buf, data_len);
+      }
     }
     else if (data_type == TYPE_BOOL || data_type == TYPE_WORD) {
       value = newSViv( buffer_get_short_le(buf) );
@@ -1107,16 +1084,16 @@ _store_tag(HV *tags, SV *key, SV *value)
   if ( my_hv_exists_ent( tags, key ) ) {
     SV **entry = my_hv_fetch( tags, SvPVX(key) );
     if (entry != NULL) {
-      // A normal string entry, convert to array.
-      if ( SvTYPE(*entry) == SVt_PV ) {
+      if ( SvTYPE(SvRV(*entry)) == SVt_PVAV ) {
+        av_push( (AV *)SvRV(*entry), value );
+      }
+      else {
+      // A non-array entry, convert to array.
         AV *ref = newAV();
         av_push( ref, newSVsv(*entry) );
         av_push( ref, value );
         my_hv_store_ent( tags, key, newRV_noinc( (SV*)ref ) );
       }
-      else if ( SvTYPE(SvRV(*entry)) == SVt_PVAV ) {
-        av_push( (AV *)SvRV(*entry), value );
-      }
     }
   }
   else {
@@ -1235,3 +1212,125 @@ _parse_index(Buffer *buf, uint64_t audio_offset, HV *info, HV *tags)
     }
   }
 }
+
+void
+_parse_content_encryption(Buffer *buf, HV *info, HV *tags)
+{
+  uint32_t protection_type_len;
+  uint32_t key_len;
+  uint32_t license_url_len;
+  
+  // Skip secret data
+  buffer_consume(buf, buffer_get_int_le(buf));
+  
+  protection_type_len = buffer_get_int_le(buf);
+  my_hv_store( info, "drm_protection_type", newSVpvn( buffer_ptr(buf), protection_type_len - 1 ) );
+  buffer_consume(buf, protection_type_len);
+  
+  key_len = buffer_get_int_le(buf);
+  my_hv_store( info, "drm_key", newSVpvn( buffer_ptr(buf), key_len - 1 ) );
+  buffer_consume(buf, key_len);
+  
+  license_url_len = buffer_get_int_le(buf);
+  my_hv_store( info, "drm_license_url", newSVpvn( buffer_ptr(buf), license_url_len - 1 ) );
+  buffer_consume(buf, license_url_len);
+}
+
+void
+_parse_script_command(Buffer *buf, HV *info, HV *tags)
+{
+  uint16_t command_count;
+  uint16_t type_count;
+  Buffer utf8_buf;
+  AV *types = newAV();
+  AV *commands = newAV();
+  
+  // Skip reserved
+  buffer_consume(buf, 16);
+  
+  command_count = buffer_get_short_le(buf);
+  type_count    = buffer_get_short_le(buf);
+  
+  while ( type_count-- ) {
+    SV *value;
+    uint16_t len = buffer_get_short_le(buf);
+    
+    buffer_get_utf16le_as_utf8(buf, &utf8_buf, len * 2);
+    value = newSVpv( buffer_ptr(&utf8_buf), 0 );
+    sv_utf8_decode(value);
+    buffer_free(&utf8_buf);
+    
+    av_push( types, value );
+  }
+  
+  while ( command_count-- ) {
+    HV *command = newHV();
+    SV *value;
+    
+    uint32_t pres_time  = buffer_get_int_le(buf);
+    uint16_t type_index = buffer_get_short_le(buf);
+    uint16_t name_len   = buffer_get_short_le(buf);
+    
+    buffer_get_utf16le_as_utf8(buf, &utf8_buf, name_len * 2);
+    value = newSVpv( buffer_ptr(&utf8_buf), 0 );
+    sv_utf8_decode(value);
+    buffer_free(&utf8_buf);
+    
+    my_hv_store( command, "time", newSVuv(pres_time) );
+    my_hv_store( command, "type", newSVuv(type_index) );
+    my_hv_store( command, "command", value );
+    
+    av_push( commands, newRV_noinc( (SV *)command ) );
+  }
+  
+  my_hv_store( info, "script_types", newRV_noinc( (SV *)types ) );
+  my_hv_store( info, "script_commands", newRV_noinc( (SV *)commands ) );
+}
+
+SV *
+_parse_picture(Buffer *buf)
+{
+  char *tmp_ptr;
+  uint16_t mime_len = 2; // to handle double-null
+  uint16_t desc_len = 2;
+  uint32_t image_len;
+  SV *mime;
+  SV *desc;
+  HV *picture = newHV();
+  Buffer utf8_buf;
+  
+  my_hv_store( picture, "image_type", newSVuv( buffer_get_char(buf) ) );
+
+  image_len = buffer_get_int_le(buf);
+  
+  // MIME type is a double-null-terminated UTF-16 string
+  tmp_ptr = buffer_ptr(buf);
+  while ( tmp_ptr[0] != '\0' || tmp_ptr[1] != '\0' ) {
+    mime_len += 2;
+    tmp_ptr += 2;
+  }
+  
+  buffer_get_utf16le_as_utf8(buf, &utf8_buf, mime_len);
+  mime = newSVpv( buffer_ptr(&utf8_buf), 0 );
+  sv_utf8_decode(mime);
+  my_hv_store( picture, "mime_type", mime );
+  buffer_free(&utf8_buf);
+  
+  // Description is a double-null-terminated UTF-16 string
+  tmp_ptr = buffer_ptr(buf);
+  while ( tmp_ptr[0] != '\0' || tmp_ptr[1] != '\0' ) {
+    desc_len += 2;
+    tmp_ptr += 2;
+  }
+  
+  buffer_get_utf16le_as_utf8(buf, &utf8_buf, desc_len);
+  desc = newSVpv( buffer_ptr(&utf8_buf), 0 );
+  sv_utf8_decode(desc);
+  my_hv_store( picture, "description", desc );
+  buffer_free(&utf8_buf);
+  
+  my_hv_store( picture, "image", newSVpvn( buffer_ptr(buf), image_len ) );
+  buffer_consume(buf, image_len);
+  
+  return newRV_noinc( (SV *)picture );
+}
\ No newline at end of file
